# -*- coding: utf-8 -*-
# Copyright (C) 2010, 2011, 2012, 2013 Sebastian Wiesner <lunaryorn@gmail.com>

# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version.

# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

"""
pyudev_gobject
==============

Glib integration.

:class:`MonitorObserver` integrates device monitoring into the Glib
mainloop by turning device events into Glib signals.

:mod:`GLib` and :mod:`GObject` must be available for import via
`gi.repository` when importing this module.

.. moduleauthor::  Sebastian Wiesner  <lunaryorn@gmail.com>

"""

from __future__ import unicode_literals

from gi.repository import GLib, GObject


class MonitorObserver(GObject.GObject):
    """An observer for device events integrating into the :mod:`GLib` mainloop.

    This class inherits :class:`~GObject.GObject` to turn device events into
    glib signals.

    >>> from pyudev import Context, Monitor
    >>> from pyudev_gobject import MonitorObserver
    >>> context = Context()
    >>> monitor = Monitor.from_netlink(context)
    >>> monitor.filter_by(subsystem='input')
    >>> observer = MonitorObserver(monitor)
    >>> def device_event(observer, device):
    ...     print('event {0} on device {1}'.format(device.action, device))
    >>> observer.connect('device-event', device_event)
    >>> monitor.start()

    """

    __gsignals__ = {
        # explicitly convert the signal to str, because glib expects the
        # *native* string type of the corresponding python version as type of
        # signal name, and str() is the name of the native string type of both
        # python versions.  We could also remove the "unicode_literals" import,
        # but I don't want to make exceptions to the standard set of future
        # imports used throughout pyudev for the sake of consistency.
        str('device-event'): (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE,
                              (GObject.TYPE_PYOBJECT,)),
    }

    def __init__(self, monitor):
        super().__init__()
        self._setup_observer(monitor)

    def _setup_observer(self, monitor):
        self.monitor = monitor
        self.event_source = None
        self.enabled = True

    @property
    def enabled(self):
        """
        Whether this observer is enabled or not.

        If ``True`` (the default), this observer is enabled, and emits events.
        Otherwise it is disabled and does not emit any events.

        """
        return self.event_source is not None

    @enabled.setter
    def enabled(self, value):
        if value and self.event_source is None:
            self.event_source = GLib.io_add_watch(
                self.monitor, GLib.IO_IN, self._process_udev_event)
        elif not value and self.event_source is not None:
            GLib.source_remove(self.event_source)
            self.event_source = None

    def _process_udev_event(self, source, condition):
        if condition == GLib.IO_IN:
            device = self.monitor.poll(timeout=0)
            if device is not None:
                self._emit_event(device)
        return True

    def _emit_event(self, device):
        self.emit('device-event', device)


GObject.type_register(MonitorObserver)



#!/usr/bin/env python
"""A systray app to set the JACK configuration from QjackCtl presets via DBus."""

import argparse
import configparser
import logging
import os
import sys

os.environ['NO_AT_BRIDGE'] = "1"  # noqa
import gi
gi.require_version('Gtk', '3.0')  # noqa
from gi.repository import Gtk, GObject

import dbus
from xdg import BaseDirectory as xdgbase

from .a2jcontrol import A2JCtlInterface
from .alsainfo import AlsaInfo
from .devmonitor import AlsaDevMonitor
from .indicator import Indicator
from .jackcontrol import (JackCfgInterface, JackCtlInterface)
from .jackselect_service import DBUS_NAME, DBUS_INTERFACE, DBUS_PATH, JackSelectService
from .qjackctlconf import get_qjackctl_presets
from .version import __version__


log = logging.getLogger('jack-select')

INTERVAL_GET_STATS = 500
INTERVAL_CHECK_CONF = 1000
INTERVAL_RESTART = 1000
DEFAULT_CONFIG = ('rncbc.org', 'QjackCtl.conf')
SETTINGS = ('jack-select', 'settings.ini')


class JackSelectApp:
    """A simple systray application to select a JACK configuration preset."""

    def __init__(self, bus=None, config=None, alsa_monitor=None, a2j_autostart=None,
                 a2j_export_hw=None, ignore_default=None):
        self.bus = bus or dbus.SessionBus()

        # load jack-select application settings
        self.load_settings()

        if alsa_monitor is None:
            self.alsa_monitor = self.app_settings.getboolean('general', 'alsa_monitor')
        else:
            self.alsa_monitor = alsa_monitor

        if ignore_default is None:
            self.ignore_default = self.app_settings.getboolean('general', 'ignore_default')
        else:
            self.ignore_default = ignore_default

        self.gui = Indicator('jack.png', "JACK-Select")
        self.gui.set_tooltip(self.tooltip_query)
        self.jack_status = {}
        self.tooltext = "No status available."

        # a2jmidi D-BUS service controller is created on-demand
        self._a2jctl = None
        self._a2j_autostart = a2j_autostart
        self._a2j_export_hw = a2j_export_hw

        if self.alsa_monitor:
            # get ALSA devices and their parameters
            self.handle_device_change(init=True)
        else:
            self.alsainfo = None

        # load QjackCtl presets
        self.qjackctl_config = config

        self.presets = None
        self.active_preset = None
        self.load_presets()

        # Create Jack control and config D-BUS interfaces
        self.dbus_connect()

        # set up periodic functions to check presets & jack status
        GObject.timeout_add(INTERVAL_CHECK_CONF, self.load_presets)
        GObject.timeout_add(INTERVAL_GET_STATS, self.get_jack_stats)
        self.jackctl.is_started(self.update_jack_status)

        # add & start DBUS service
        self.dbus_service = JackSelectService(self, bus)

        if self.alsa_monitor:
            # set up udev device monitor
            self.alsadevmonitor = AlsaDevMonitor(self.handle_device_change)
            self.alsadevmonitor.start()

    def dbus_connect(self):
        """Create Jack control and config D-BUS interfaces."""
        try:
            log.debug("Connecting to JACK D-BUS interface...")
            self.jackctl = JackCtlInterface(bus=self.bus)
            self.jackcfg = JackCfgInterface(bus=self.bus)
        except dbus.exceptions.DBusException as exc:
            log.warning("Could not connect to JACK D-BUS interface: %s", exc)
            return True
        else:
            log.debug("JACK D-BUS connection established.")
            self.jackctl.add_signal_handler(self.handle_jackctl_signal)

    @property
    def a2jctl(self):
        if self._a2jctl is None:
            try:
                self._a2jctl = A2JCtlInterface(bus=self.bus)
            except dbus.DBusException as exc:
                log.warning("Could not connect to a2jmidid D-BUS service.")
                log.debug("D-Bus exception: %s", exc)
            else:
                self._a2jctl.add_signal_handler(self.handle_a2jctl_signal)

        return self._a2jctl

    @property
    def a2j_autostart(self):
        return (self.app_settings.getboolean('a2jmidi', 'autostart')
                if self._a2j_autostart is None else self._a2j_autostart)

    @a2j_autostart.setter
    def a2j_autostart(self, flag):
        if flag != self.app_settings.getboolean('a2jmidi', 'autostart'):
            self.app_settings.set('a2jmidi', 'autostart', 'yes' if flag else 'no')
            self.write_settings()

    @property
    def a2j_export_hw(self):
        return (self.app_settings.getboolean('a2jmidi', 'export_hw')
                if self._a2j_export_hw is None else self._a2j_export_hw)

    @a2j_export_hw.setter
    def a2j_export_hw(self, flag):
        if flag != self.app_settings.getboolean('a2jmidi', 'export_hw'):
            self.app_settings.set('a2jmidi', 'export_hw', 'yes' if flag else 'no')
            self.write_settings()

    def load_settings(self):
        self.app_settings = configparser.ConfigParser()
        self.app_settings.read_dict({
            'general': {
                'alsa_monitor': 'yes',
                'ignore_default': 'no',
            },
            'a2jmidi': {
                'autostart': 'no',
                'export_hw': 'yes',
            }
        })

        settings_file = xdgbase.load_first_config(*SETTINGS)

        if settings_file:
            log.debug("Loading settings from '%s'.", settings_file)
            self.app_settings.read(settings_file)

    def write_settings(self):
        try:
            settings_file = os.path.join(xdgbase.save_config_path(SETTINGS[0]), SETTINGS[1])
            log.debug("Writing settings to '%s'.", settings_file)
            with open(settings_file, 'w') as fp:
                self.app_settings.write(fp)
        except OSError as exc:
            log.error("Could not write settings file '%s': %s", settings_file, exc)

    def load_presets(self, force=False):
        if self.qjackctl_config in (None, DEFAULT_CONFIG):
            qjackctl_config = xdgbase.load_first_config(*DEFAULT_CONFIG)
        else:
            if os.access(self.qjackctl_config, os.R_OK):
                qjackctl_config = self.qjackctl_config
            else:
                qjackctl_config = None

        if qjackctl_config:
            mtime = os.path.getmtime(qjackctl_config)
            changed = mtime > getattr(self, '_conf_mtime', 0)

            if changed:
                log.debug("JACK configuration file mtime changed or previously unknown.")

            if force or changed or self.presets is None:
                log.debug("(Re-)Reading configuration.")
                (
                    preset_names,
                    self.jack_settings,
                    self.default_preset
                ) = get_qjackctl_presets(qjackctl_config, self.ignore_default)
                self.presets = {name: name.replace('_', ' ')
                                for name in preset_names}
                self.create_menu()

            self._conf_mtime = mtime
        elif self.presets or self.presets is None:
            log.warning("Could not access configuration file.")

            if __debug__ and self.presets:
                log.debug("Removing stored presets from memory.")

            self.presets = {}
            self.jack_settings = {}
            self.default_preset = None
            self.create_menu()

        return True  # keep function scheduled

    def check_alsa_settings(self, preset):
        engine = self.jack_settings[preset]['engine']
        driver = self.jack_settings[preset]['driver']
        if engine['driver'] != 'alsa':
            return True

        dev = driver.get('device')
        if dev and dev not in self.alsainfo.devices:
            log.debug("Device '%s' used by preset '%s' not found.",
                      dev, preset)
            return False

        dev = driver.get('playback')
        if dev and dev not in self.alsainfo.playback_devices:
            log.debug("Playback device '%s' used by preset '%s' not found.",
                      dev, preset)
            return False

        dev = driver.get('capture')
        if dev and dev not in self.alsainfo.capture_devices:
            log.debug("Capture device '%s' used by preset '%s' not found.",
                      dev, preset)
            return False

        return True

    def create_menu(self):
        log.debug("Building menu.")
        self.gui.clear_menu()

        if self.presets:
            if not self.alsainfo:
                log.debug("ALSA device info not available. Filtering disabled.")

            callback = self.activate_preset
            for name, label in sorted(self.presets.items()):
                disabled = self.alsainfo and not self.check_alsa_settings(name)
                self.gui.add_menu_item(callback, label, enabled=not disabled, data=name)

        else:
            self.gui.add_menu_item(None, "No presets found", enabled=False)

        self.gui.add_separator()
        self.menu_stop = self.gui.add_menu_item(self.stop_jack_server,
                                                "Stop JACK Server",
                                                icon='stop.png',
                                                enabled=bool(self.jack_status.get('is_started')))

        if self.a2jctl:
            self.gui.add_separator()
            self.menu_a2jbridge = self.gui.add_submenu('ALSA-MIDI Bridge')
            self.menu_a2j_startstop = self.gui.add_menu_item(self.on_start_stop_a2jbridge,
                                                             "ALSA-MIDI Bridge",
                                                             icon='midi.png',
                                                             menu=self.menu_a2jbridge)
            self.menu_a2j_export_hw = self.gui.add_menu_item(self.on_a2jbridge_set_export_hw,
                                                             "Export HW Ports",
                                                             is_check=True,
                                                             active=self.a2j_export_hw,
                                                             menu=self.menu_a2jbridge)
            self.menu_a2j_autostart = self.gui.add_menu_item(self.on_a2jbridge_autostart,
                                                             "Auto-Start with JACK",
                                                             is_check=True,
                                                             active=self.a2j_autostart,
                                                             menu=self.menu_a2jbridge)
        else:
            self.menu_a2jbridge = None

        self.gui.add_separator()
        self.menu_quit = self.gui.add_menu_item(self.quit, "Quit", icon='quit.png')
        self.gui.menu.show_all()
        self.update_a2jbridge_status()

    def open_menu(self):
        self.gui.on_popup_menu_open()

    def update_jack_status(self, value, name=None):
        jack_started = self.jack_status.get('is_started')
        self.jack_status[name] = value

        if name == 'is_started' and value != jack_started:
            if value:
                self.gui.set_icon('started.png')
                log.info("JACK server has started.")
                self.a2jbridge_autostart()
            else:
                self.gui.set_icon('stopped.png')
                self.tooltext = "JACK server is stopped."
                log.info(self.tooltext)

            self.menu_stop.set_sensitive(value)
            self.update_a2jbridge_status()

        if self.jack_status.get('is_started'):
            try:
                if self.active_preset:
                    label = self.presets.get(self.active_preset,
                                             self.active_preset)
                    self.tooltext = "<b>[%s]</b>\n" % label
                else:
                    self.tooltext = "<i><b>Unknown configuration</b></i>\n"

                self.tooltext += ("%(samplerate)i Hz / %(period)i frames "
                                  "(%(latency)0.1f ms)\n" % self.jack_status)
                self.tooltext += "RT: %s " % (
                    "yes" if self.jack_status.get('is_realtime') else "no")
                self.tooltext += ("load: %(load)i%% xruns: %(xruns)i" %
                                  self.jack_status)
            except KeyError:
                self.tooltext = "No status available."

    def update_a2jbridge_status(self, status=None):
        if self.menu_a2jbridge:
            if not self.a2jctl:
                # No a2jmidid service D-BUS interface
                self.menu_a2j_startstop.set_sensitive(False)
                self.menu_a2j_export_hw.set_sensitive(False)
                self.menu_a2j_startstop.set_label("ALSA-MIDI Bridge not available")
            elif self.jack_status.get('is_started'):
                # JACK server started
                if status is None:
                    status = self.a2jctl.is_started()

                if status:
                    # bridge started
                    self.menu_a2j_startstop.set_label("Stop ALSA-MIDI Bridge")
                    self.menu_a2j_export_hw.set_active(self.a2jctl.get_hw_export())
                    self.menu_a2j_export_hw.set_sensitive(False)
                else:
                    # bridge stopped
                    self.menu_a2j_startstop.set_label("Start ALSA-MIDI Bridge")
                    self.menu_a2j_export_hw.set_sensitive(True)

                self.menu_a2j_startstop.set_sensitive(True)
            else:
                # JACK server stopped
                self.menu_a2j_startstop.set_label("ALSA-MIDI Bridge suspended")
                self.menu_a2j_startstop.set_sensitive(False)
                self.menu_a2j_export_hw.set_sensitive(True)

    def handle_device_change(self, observer=None, device=None, init=False):
        if device:
            dev = device.device_path.split('/')[-1]

        if init or (device.action in ('change', 'remove')
                    and dev.startswith('card')):
            try:
                log.debug("Sound device change signalled. Collecting ALSA "
                          "device info...")
                self.alsainfo = AlsaInfo(deferred=False)
            except Exception as exc:
                log.warn("Could not get ALSA device list: %s", exc)
                self.alsainfo = None

            if device and device.action != 'init':
                self.load_presets(force=True)

    def handle_jackctl_signal(self, *args, signal=None, **kw):
        log.debug("JackCtl signal received: %r", signal)
        if signal == 'ServerStarted':
            self.update_jack_status(True, name='is_started')
        elif signal == 'ServerStopped':
            self.update_jack_status(False, name='is_started')

    def handle_a2jctl_signal(self, *args, signal=None, **kw):
        if signal == 'bridge_started':
            log.debug("a2jmidid bridge STARTED signal received.")
            self.update_a2jbridge_status(True)
        elif signal == 'bridge_stopped':
            log.debug("a2jmidid bridge STOPPED signal received.")
            self.update_a2jbridge_status(False)

    def handle_dbus_error(self, *args):
        """Handle errors from async JackCtlInterface calls.

        If the error indicates that the JackCtl D-BUS service vanished,
        invalidate the existing D-BUS interface instances and schedule a
        reconnection attempt.

        """
        log.warning("JackCtl D-BUS call error handler called.")
        if args and isinstance(args[0], dbus.DBusException):
            if 'org.freedesktop.DBus.Error.ServiceUnknown' in str(args[0]) and self.jackctl:
                log.warning("JackCtl D-BUS service vanished. Assuming JACK is stopped.")
                self.update_jack_status(False, name='is_started')
                self.jackctl = None
                self.jackcfg = None
                GObject.timeout_add(INTERVAL_GET_STATS, self.dbus_connect)

    def get_jack_stats(self):
        if self.jackctl and self.jack_status.get('is_started'):
            try:
                cb = self.update_jack_status
                ecb = self.handle_dbus_error
                self.jackctl.is_realtime(cb, ecb)
                self.jackctl.get_sample_rate(cb, ecb)
                self.jackctl.get_period(cb, ecb)
                self.jackctl.get_load(cb, ecb)
                self.jackctl.get_xruns(cb, ecb)
                self.jackctl.get_latency(cb, ecb)
            except dbus.exceptions.DBusException:
                log.warning("JackCtl D-BUS service failure. Assuming JACK is stopped.")
                self.update_jack_status(False, name='is_started')

        return True  # keep function scheduled

    def tooltip_query(self, widget, x, y, keyboard_mode, tooltip):
        """Set tooltip for the systray icon."""
        if self.jackctl:
            tooltip.set_markup(self.tooltext)
        else:
            tooltip.set_text("No JACK-DBus connection")

        return True

    def activate_default_preset(self):
        if self.default_preset:
            log.debug("Activating default preset '%s'.", self.default_preset)
            self.activate_preset(preset=self.default_preset)
        else:
            log.warn("No default preset set.")

    def activate_preset(self, m_item=None, **kwargs):
        if m_item:
            preset = m_item.data
        else:
            preset = kwargs.get('preset')

        if not preset:
            log.warn("Preset must not be null.")
            return

        settings = self.jack_settings.get(preset)

        if settings:
            if self.jackcfg:
                self.jackcfg.activate_preset(settings)
                log.info("Activated preset: %s", preset)
                self.stop_jack_server()
                GObject.timeout_add(INTERVAL_RESTART, self.start_jack_server)
                self.active_preset = preset
        else:
            log.error("Unknown preset '%s'. Ignoring it.", preset)

    def start_jack_server(self, *args, **kwargs):
        if self.jackctl and not self.jack_status.get('is_started'):
            log.debug("Starting JACK server...")
            try:
                self.jackctl.start_server()
            except Exception as exc:
                log.error("Could not start JACK server: %s", exc)

    def stop_jack_server(self, *args, **kwargs):
        if self.jackctl and self.jack_status.get('is_started'):
            self.active_preset = None
            log.debug("Stopping JACK server...")

            try:
                self.jackctl.stop_server()
            except Exception as exc:
                log.error("Could not stop JACK server: %s", exc)

    def on_start_stop_a2jbridge(self, *args):
        self.start_stop_a2jbridge()

    def start_stop_a2jbridge(self, start_stop=None):
        if not self.a2jctl:
            return

        if start_stop is None:
            start_stop = not self.a2jctl.is_started()

        if start_stop:
            log.debug("Export HW ports: %s", "yes" if self.a2j_export_hw else "no")
            self.a2jctl.set_hw_export(self.a2j_export_hw)

            log.debug("Starting ALSA-MIDI to JACK bridge...")
            self.a2jctl.start()
        else:
            log.debug("Stopping ALSA-MIDI to JACK bridge...")
            self.a2jctl.stop()

    def on_a2jbridge_set_export_hw(self, widget, *args):
        self.a2j_export_hw = widget.get_active()
        log.debug("a2jmidid hw export %sabled.", 'en' if self.a2j_autostart else 'dis')

    def on_a2jbridge_autostart(self, widget, *args):
        self.a2j_autostart = widget.get_active()
        log.debug("a2jmidid auto-start %sabled.", 'en' if self.a2j_autostart else 'dis')

    def a2jbridge_autostart(self):
        if self.a2jctl and self.a2j_autostart and not self.a2jctl.is_started():
            log.debug("a2jmidid auto-start triggered.")
            self.start_stop_a2jbridge(True)

    def quit(self, *args):
        log.debug("Exiting main loop.")
        Gtk.main_quit()


def get_dbus_client(bus=None):
    if bus is None:
        bus = dbus.SessionBus()

    obj = bus.get_object(DBUS_NAME, DBUS_PATH)
    return dbus.Interface(obj, DBUS_INTERFACE)


def main(args=None):
    """Main function to be used when called as a script."""
    from dbus.mainloop.glib import DBusGMainLoop

    ap = argparse.ArgumentParser(prog='jack-select', description=__doc__.splitlines()[0])
    ap.add_argument(
        '--version',
        action="version",
        version="%%(prog)s %s" % __version__,
        help="Show program version and exit.")
    ap.add_argument(
        '--a2j-autostart',
        action="store_true",
        default=None,
        help="Autostart ALSA-MIDI to JACK bridge with JACK.")
    ap.add_argument(
        '--a2j-export-hw',
        action="store_true",
        default=None,
        help="Export hardware MIDI ports via ALSA-MIDI to JACK bridge.")
    ap.add_argument(
        '-a', '--no-alsa-monitor',
        action="store_false",
        default=None,
        help="Disable ALSA device monitoring and filtering.")
    ap.add_argument(
        '-c', '--config',
        metavar='PATH',
        help="Path to configuration file (default: <XDG_CONFIG_HOME>/%s/%s)" % DEFAULT_CONFIG)
    ap.add_argument(
        '-d', '--default',
        action="store_true",
        help="Activate default preset.")
    ap.add_argument(
        '-i', '--ignore-default',
        action="store_true",
        default=None,
        help="Ignore the nameless '(default)' preset if any other presets are stored in the "
             "configuration.")
    ap.add_argument(
        '-v', '--verbose',
        action="store_true",
        help="Be verbose about what the script does.")
    ap.add_argument(
        'preset',
        nargs='?',
        help="Configuration preset to activate on startup.")

    args = ap.parse_args(args if args is not None else sys.argv[1:])

    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO,
                        format="[%(name)s] %(levelname)s: %(message)s")

    # the mainloop needs to be set before creating the session bus instance
    DBusGMainLoop(set_as_default=True)
    bus = dbus.SessionBus()
    start_gui = False

    try:
        client = get_dbus_client(bus)
        log.debug("JACK-Select DBus service detected.")

        if args.default:
            log.debug("Activating default preset.")
            client.ActivateDefaultPreset()
        elif args.preset:
            log.debug("Activating preset '%s'.", args.preset)
            client.ActivatePreset(args.preset)
        else:
            log.debug("Opening menu...")
            client.OpenMenu()
    except dbus.DBusException as exc:
        if exc.get_dbus_name().endswith('ServiceUnknown'):
            start_gui = True
        else:
            log.warning("Exception: %s", exc)

    log.debug("Args: %s", args)

    if start_gui:
        app = JackSelectApp(bus,
                            config=args.config,
                            a2j_autostart=args.a2j_autostart,
                            a2j_export_hw=args.a2j_export_hw,
                            alsa_monitor=args.no_alsa_monitor,
                            ignore_default=args.ignore_default)

        if args.default:
            # load default preset when mainloop starts
            GObject.timeout_add(0, app.activate_default_preset)
        elif args.preset:
            # load given preset when mainloop starts
            GObject.timeout_add(0, lambda: app.activate_preset(preset=args.preset))

        try:
            return Gtk.main()
        except KeyboardInterrupt:
            return "Interrupted."


if __name__ == '__main__':
    sys.exit(main() or 0)



# -*- coding: utf-8 -*-
"""Set up an udev monitor to be notified about changes in attached sound devices."""

import logging

from pyudev import Context, Monitor
from .pyudev_gobject import MonitorObserver


log = logging.getLogger(__name__)


class AlsaDevMonitor:
    def __init__(self, callback):
        # set up udev device monitor
        context = Context()
        self._monitor = Monitor.from_netlink(context)
        self._monitor.filter_by(subsystem='sound')
        self._observer = MonitorObserver(self._monitor)
        self._observer.connect('device-event', callback)

    def start(self):
        log.debug("Starting AlsaDevMonitor...")
        self._monitor.start()



"""A convenience class for a GTK 3 system tray indicator."""

from pkg_resources import resource_filename

import gi
gi.require_version('Gtk', '3.0')  # noqa
from gi.repository import Gtk
from gi.repository.GdkPixbuf import Pixbuf


class Indicator:
    """This class defines a standard GTK3 system tray indicator.

    Class Indicator can be easily reused in any other project.

    """
    def __init__(self, icon, title=None):
        """Create indicator icon and add menu.

        Args:
          icon (str): path to initial icon that will be shown on system panel

        """
        self._icon_cache = {}
        self.icon = Gtk.StatusIcon.new_from_pixbuf(self._get_icon(icon))
        self.menu = Gtk.Menu()
        self.icon.connect('activate', self.on_popup_menu_open)
        self.icon.connect('popup-menu', self.on_popup_menu_open)

        if title:
            self.icon.set_title(title)

    def _get_icon(self, icon):
        """Return icon from package as GdkPixbuf.Pixbuf.

        Extracts the image from package to a file, stores it in the icon cache
        if it's not in there yet and returns it. Otherwise just returns the
        image stored in the cache.

        """
        if icon not in self._icon_cache:
            filename = resource_filename(__name__, "images/%s" % icon)
            self._icon_cache[icon] = Pixbuf.new_from_file(filename)

        return self._icon_cache[icon]

    def set_icon(self, icon):
        """Set new icon in system tray.

        Args:
          icon (str): path to file with new icon

        """
        self.icon.set_from_pixbuf(self._get_icon(icon))

    def set_tooltip(self, callback):
        self.icon.set_has_tooltip(True)
        self.icon.connect("query-tooltip", callback)

    def clear_menu(self):
        """Clear all entries from the main menu."""
        self.menu = Gtk.Menu()

    def add_menu_item(self, command=None, title=None, icon=None, enabled=True, is_check=False,
                      active=False, menu=None, data=None):
        """Add mouse right click menu item.

        Args:
          command (callable): function that will be called after left mouse
          click on title
          title (str): label that will be shown in menu
          icon (str): name of icon stored in application package
          active (bool): whether the menu entry can be activated (default: True)
          data (obj): arbitrary data to associate with the menu entry

        """
        if icon:
            m_item = Gtk.ImageMenuItem(title)
            image = Gtk.Image.new_from_pixbuf(self._get_icon(icon))
            m_item.set_image(image)
        elif is_check:
            m_item = Gtk.CheckMenuItem(title)
            m_item.set_active(active)
        else:
            m_item = Gtk.MenuItem(title)

        if command:
            m_item.connect('toggled' if is_check else 'activate', command)

        m_item.set_sensitive(enabled)
        m_item.data = data

        if menu:
            menu.append(m_item)
        else:
            self.menu.append(m_item)

        return m_item

    def add_submenu(self, title):
        """Add a sub menu popup menu."""
        submenu = Gtk.Menu()
        m_item = Gtk.MenuItem(title)
        m_item.set_submenu(submenu)
        self.menu.append(m_item)
        return submenu

    def add_separator(self):
        """Add separator between labels in the popup menu."""
        m_item = Gtk.SeparatorMenuItem()
        self.menu.append(m_item)

    def on_popup_menu_open(self, widget=None, button=None, *args):
        """Some action requested opening the popup menu."""
        self.menu.popup(None, None, Gtk.StatusIcon.position_menu,
                        widget or self.icon, button or 1,
                        Gtk.get_current_event_time())

    def on_popup_menu_close(self, widget=None, button=None, *args):
        """Some action requested closing the popup menu."""
        self.menu.popdown()



# -*- coding: utf-8 -*-
"""Base class for D-BUS service interface wrappers."""

import logging
from functools import partial

import dbus


log = logging.getLogger(__name__)


class DBUSBaseInterface:
    """Base class for D-BUS service interface wrappers.

    This is an abstract base class. Sub-classes ned to define the following
    class or instance attributes:

    * service - the name of the D-BUS service
    * interface - the name of the D-BUS accessed via this class
    * object_path = the path to the service object providing the interface

    """

    def __init__(self, ctl=None, bus=None):
        if not ctl:
            ctl = self.get_controller(bus)

        self._if = dbus.Interface(ctl, self.interface)

    def get_controller(self, bus=None):
        if not bus:
            bus = dbus.SessionBus()
        return bus.get_object(self.service, self.object_path)

    def _async_handler(self, *args, **kw):
        name = kw.get('name')
        callback = kw.get('callback')

        if args and isinstance(args[0], dbus.DBusException):
            log.error("Async call failed name=%s: %s", name, args[0])
            return

        if callback:
            callback(*args, name=name)

    def call_async(self, meth, args=None, name=None, callback=None,
                   error_callback=None):
        if callback:
            handler = partial(self._async_handler, callback=callback,
                              name=name or meth)
            kw = dict(reply_handler=handler, error_handler=error_callback or handler)
        else:
            kw = {}

        return getattr(self._if, meth)(*args or [], **kw)

    def add_signal_handler(self, handler, signal=None):
        return self._if.connect_to_signal(
            signal_name=signal,
            handler_function=handler,
            interface_keyword='interface',
            member_keyword='signal')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from collections import namedtuple
from ctypes import c_char_p, c_int, c_uint, c_ulong, c_void_p, byref, cdll, create_string_buffer
from enum import IntEnum
from textwrap import indent


log = logging.getLogger(__name__)


class LibAsoundError(Exception):
    pass


class SndPcmStream(IntEnum):
    PLAYBACK = 0
    CAPTURE = 1


class SndPcmFormat(IntEnum):
    # Signed 8 bit
    S8 = 0
    # Unsigned 8 bit
    U8 = 1
    # Signed 16 bit Little Endian
    S16_LE = 2
    # Signed 16 bit Big Endian
    S16_BE = 3
    # Unsigned 16 bit Little Endian
    U16_LE = 4
    # Unsigned 16 bit Big Endian
    U16_BE = 5
    # Signed 24 bit Little Endian using low three bytes in 32-bit word
    S24_LE = 6
    # Signed 24 bit Big Endian using low three bytes in 32-bit word
    S24_BE = 7
    # Unsigned 24 bit Little Endian using low three bytes in 32-bit word
    U24_LE = 8
    # Unsigned 24 bit Big Endian using low three bytes in 32-bit word
    U24_BE = 9
    # Signed 32 bit Little Endian
    S32_LE = 10
    # Signed 32 bit Big Endian
    S32_BE = 11
    # Unsigned 32 bit Little Endian
    U32_LE = 12
    # Unsigned 32 bit Big Endian
    U32_BE = 13
    # Float 32 bit Little Endian, Range -1.0 to 1.0
    FLOAT_LE = 14
    # Float 32 bit Big Endian, Range -1.0 to 1.0
    FLOAT_BE = 15
    # Float 64 bit Little Endian, Range -1.0 to 1.0
    FLOAT64_LE = 16
    # Float 64 bit Big Endian, Range -1.0 to 1.0
    FLOAT64_BE = 17
    # IEC-958 Little Endian
    IEC958_SUBFRAME_LE = 18
    # IEC-958 Big Endian
    IEC958_SUBFRAME_BE = 19
    # Mu-Law
    MU_LAW = 20
    # A-Law
    A_LAW = 21
    # Ima-ADPCM
    IMA_ADPCM = 22
    # MPEG
    MPEG = 23
    # GSM
    GSM = 24
    # Signed 20bit Little Endian in 4bytes format, LSB justified
    S20_LE = 25
    # Signed 20bit Big Endian in 4bytes format, LSB justified
    S20_BE = 26
    # Unsigned 20bit Little Endian in 4bytes format, LSB justified
    U20_LE = 27
    # Unsigned 20bit Big Endian in 4bytes format, LSB justified
    U20_BE = 28
    # Special
    SPECIAL = 31
    # Signed 24bit Little Endian in 3bytes format
    S24_3LE = 32
    # Signed 24bit Big Endian in 3bytes format
    S24_3BE = 33
    # Unsigned 24bit Little Endian in 3bytes format
    U24_3LE = 34
    # Unsigned 24bit Big Endian in 3bytes format
    U24_3BE = 35
    # Signed 20bit Little Endian in 3bytes format
    S20_3LE = 36
    # Signed 20bit Big Endian in 3bytes format
    S20_3BE = 37
    # Unsigned 20bit Little Endian in 3bytes format
    U20_3LE = 38
    # Unsigned 20bit Big Endian in 3bytes format
    U20_3BE = 39
    # Signed 18bit Little Endian in 3bytes format
    S18_3LE = 40
    # Signed 18bit Big Endian in 3bytes format
    S18_3BE = 41
    # Unsigned 18bit Little Endian in 3bytes format
    U18_3LE = 42
    # Unsigned 18bit Big Endian in 3bytes format
    U18_3BE = 43
    # G.723 (ADPCM) 24 kbit/s, 8 samples in 3 bytes
    G723_24 = 44
    # G.723 (ADPCM) 24 kbit/s, 1 sample in 1 byte
    G723_24_1B = 45
    # G.723 (ADPCM) 40 kbit/s, 8 samples in 3 bytes
    G723_40 = 46
    # G.723 (ADPCM) 40 kbit/s, 1 sample in 1 byte
    G723_40_1B = 47
    # Direct Stream Digital (DSD) in 1-byte samples (x8)
    DSD_U8 = 48
    # Direct Stream Digital (DSD) in 2-byte samples (x16)
    DSD_U16_LE = 49
    # Direct Stream Digital (DSD) in 4-byte samples (x32)
    DSD_U32_LE = 50
    # Direct Stream Digital (DSD) in 2-byte samples (x16)
    DSD_U16_BE = 51
    # Direct Stream Digital (DSD) in 4-byte samples (x32)
    DSD_U32_BE = 52

    if sys.byteorder == 'little':
        # Signed 16 bit CPU endian
        S16 = S16_LE
        # Unsigned 16 bit CPU endian
        U16 = U16_LE
        # Signed 24 bit CPU endian
        S24 = S24_LE
        # Unsigned 24 bit CPU endian
        U24 = U24_LE
        # Signed 32 bit CPU endian
        S32 = S32_LE
        # Unsigned 32 bit CPU endian
        U32 = U32_LE
        # Float 32 bit CPU endian
        FLOAT = FLOAT_LE
        # Float 64 bit CPU endian
        FLOAT64 = FLOAT64_LE
        # IEC-958 CPU Endian
        IEC958_SUBFRAME = IEC958_SUBFRAME_LE
        # Signed 20bit in 4bytes format, LSB justified, CPU Endian
        S20 = S20_LE
        # Unsigned 20bit in 4bytes format, LSB justified, CPU Endian
        U20 = U20_LE
    elif sys.byteorder == 'big':
        # Signed 16 bit CPU endian
        S16 = S16_BE
        # Unsigned 16 bit CPU endian
        U16 = U16_BE
        # Signed 24 bit CPU endian
        S24 = S24_BE
        # Unsigned 24 bit CPU endian
        U24 = U24_BE
        # Signed 32 bit CPU endian
        S32 = S32_BE
        # Unsigned 32 bit CPU endian
        U32 = U32_BE
        # Float 32 bit CPU endian
        FLOAT = FLOAT_BE
        # Float 64 bit CPU endian
        FLOAT64 = FLOAT64_BE
        # IEC-958 CPU Endian
        IEC958_SUBFRAME = IEC958_SUBFRAME_BE
        # Signed 20bit in 4bytes format, LSB justified, CPU Endian
        S20 = S20_BE
        # Unsigned 20bit in 4bytes format, LSB justified, CPU Endian
        U20 = U20_BE


PCM_RATES = (
    5512,
    8000,
    11025,
    16000,
    22050,
    32000,
    44100,
    48000,
    64000,
    88200,
    96000,
    176400,
    192000
)
PCM_BUFFER_SIZES = (
    32,
    64,
    128,
    256,
    512,
    1024,
    2048,
    4096
)
SND_PCM_NONBLOCK = 1

_lib = cdll.LoadLibrary("libasound.so.2")
_lib.snd_ctl_card_info_get_id.restype = c_char_p
_lib.snd_ctl_card_info_get_name.restype = c_char_p
_lib.snd_pcm_hw_params_get_format_mask.restype = None
_lib.snd_pcm_info_get_id.restype = c_char_p
_lib.snd_pcm_info_get_name.restype = c_char_p
_lib.snd_pcm_info_get_subdevice_name.restype = c_char_p
_lib.snd_pcm_format_name.restype = c_char_p
_lib.snd_strerror.restype = c_char_p


def decode_format_mask(fmask):
    return ((fmt, _lib.snd_pcm_format_name(c_int(fmt)).decode())
            for fmt in SndPcmFormat
            if _lib.snd_pcm_format_mask_test(fmask, c_int(fmt)) > 0)


def check_call(fn, args, msg="{errmsg}", **kwargs):
    err = fn(*args)
    if fn.restype is not None and err < 0:
        if '{errmsg}' not in msg:
            msg += ' {errmsg}'
        errmsg = _lib.snd_strerror(err).decode('utf-8')
        raise LibAsoundError(msg.format(errmsg=errmsg, **kwargs))


class AlsaCard(namedtuple('AlsaCard', ('cardno', 'id', 'name', 'devices'))):
    __slots__ = ()

    def __repr__(self):
        s = 'AlsaCard(cardno={c.cardno}, id={c.id!r}, name={c.name!r} devices=['.format(c=self)
        if self.devices:
            s += '\n' + indent(',\n'.join(repr(d) for d in self.devices), ' ' * 4)
            s += '\n'

        s += '])'
        return s


class AlsaDevice(namedtuple('AlsaDevice', ('devno', 'id', 'name', 'stream', 'buffer_sizes',
                                           'periods', 'channels', 'rates', 'formats',
                                           'subdevices'))):
    __slots__ = ()

    def __repr__(self):
        s = 'AlsaDevice(\n'
        for name, value in self._asdict().items():
            if name == 'stream':
                value = str(value)
            elif name == 'formats' and value is not None:
                value = tuple(f[1] for f in value)

            s += '    {}: {},\n'.format(name, value)

        s += ')'
        return s


def get_cards(stream=SndPcmStream.PLAYBACK):
    if stream not in SndPcmStream:
        raise Exception("Unknown stream type: {}".format(stream))

    cards = []
    c_card = c_int(-1)
    c_dev = c_int(-1)
    c_dir = c_int(0)
    c_min = c_uint()
    c_max = c_uint()
    c_min_long = c_ulong()
    c_max_long = c_ulong()
    c_handle_p = c_void_p()
    c_pcm_p = c_void_p()
    c_info_p = c_void_p()
    c_pcminfo_p = c_void_p()
    c_fmask_p = c_void_p()
    c_params_p = c_void_p()

    s_stream = 'playback' if stream == SndPcmStream.PLAYBACK else 'capture'

    # card enumeration
    while True:
        _lib.snd_card_next(byref(c_card))
        if c_card.value < 0:
            log.debug("End of card enumeration list reached.")
            break

        hwdev = "hw:{}".format(c_card.value)
        b_hwdev = create_string_buffer(hwdev.encode())

        err = _lib.snd_ctl_open(byref(c_handle_p), b_hwdev, c_card)
        if err < 0:
            _lib.snd_ctl_close(c_handle_p)
            continue

        check_call(_lib.snd_ctl_card_info_malloc, (byref(c_info_p),),
                   "Could not allocate memory for snd_ctl_card_info_t.")
        _lib.snd_ctl_card_info(c_handle_p, c_info_p)

        devices = []
        card = AlsaCard(
            cardno=c_card.value,
            id=_lib.snd_ctl_card_info_get_id(c_info_p).decode(),
            name=_lib.snd_ctl_card_info_get_name(c_info_p).decode(),
            devices=devices
        )
        log.debug('Discovered card #%i "%s" ("%s").', card.cardno, card.id, card.name)
        cards.append(card)

        # device enumeration
        while True:
            _lib.snd_ctl_pcm_next_device(c_handle_p, byref(c_dev))
            if c_dev.value < 0:
                log.debug("End of device enumeration list reached.")
                break

            check_call(_lib.snd_pcm_info_malloc, (byref(c_pcminfo_p),),
                       "Could not allocate memory for snd_pcm_info_t.")
            _lib.snd_pcm_info_set_device(c_pcminfo_p, c_dev)
            _lib.snd_pcm_info_set_subdevice(c_pcminfo_p, 0)
            _lib.snd_pcm_info_set_stream(c_pcminfo_p, c_int(stream))

            err = _lib.snd_ctl_pcm_info(c_handle_p, c_pcminfo_p)
            if err < 0:
                errmsg = _lib.snd_strerror(err).decode('utf-8')
                log.debug("Could not get info for PCM %s device #%i. %s",
                          s_stream, c_dev.value, errmsg)
                continue

            device_id = bytes.decode(_lib.snd_pcm_info_get_id(c_pcminfo_p))
            device_name = bytes.decode(_lib.snd_pcm_info_get_name(c_pcminfo_p))
            log.debug('Discovered %s device #%i "%s" ("%s").', s_stream, c_dev.value, device_id,
                      device_name)

            # count subdevices
            nsubd = _lib.snd_pcm_info_get_subdevices_count(c_pcminfo_p)
            log.debug("Device has %i subdevice(s).", nsubd)

            subdevices = []
            # open sound device
            hwdev = "hw:{},{}".format(card.id, c_dev.value)
            b_hwdev = create_string_buffer(hwdev.encode('ascii'))
            buffer_sizes = periods = channels = rates = formats = None

            try:
                check_call(_lib.snd_pcm_open,
                           (byref(c_pcm_p), b_hwdev, c_int(stream), SND_PCM_NONBLOCK),
                           "Could not open PCM {stream} device '{dev}'.",
                           stream=s_stream, dev=hwdev)
                check_call(_lib.snd_pcm_nonblock, (c_pcm_p, 1),
                           "Nonblock setting error: ")
            except LibAsoundError as exc:
                log.warning(str(exc))
            else:
                try:
                    # Get hardware parameter space
                    check_call(_lib.snd_pcm_hw_params_malloc, (byref(c_params_p),),
                               "Could not allocate memory for snd_pcm_hw_params_t.")
                    check_call(_lib.snd_pcm_hw_params_any, (c_pcm_p, c_params_p),
                               "Could not get params for {stream} device '{dev}'.",
                               stream=s_stream, dev=hwdev)

                    # Get supported channel counts
                    check_call(_lib.snd_pcm_hw_params_get_channels_min, (c_params_p, byref(c_min)),
                               "Could not get minimum channels count.")

                    check_call(_lib.snd_pcm_hw_params_get_channels_max, (c_params_p, byref(c_max)),
                               "Could not get maximum channels count.")

                    log.debug("Min/max channels: %i, %i", c_min.value, c_max.value)
                    channels = tuple(
                        ch for ch in range(c_min.value, c_max.value + 1)
                        if _lib.snd_pcm_hw_params_test_channels(c_pcm_p, c_params_p, ch) == 0
                    )

                    # Get supported sample rates
                    check_call(_lib.snd_pcm_hw_params_get_rate_min,
                               (c_params_p, byref(c_min), byref(c_dir)),
                               "Could not get minimum sample rate.")

                    check_call(_lib.snd_pcm_hw_params_get_rate_max,
                               (c_params_p, byref(c_max), byref(c_dir)),
                               "Could not get maximum sample rate.")

                    log.debug("Min/max sample rate: %i, %i", c_min.value, c_max.value)
                    rates = tuple(
                        rate for rate in PCM_RATES
                        if c_min.value <= rate <= c_max.value and
                        _lib.snd_pcm_hw_params_test_rate(c_pcm_p, c_params_p, rate, 0) == 0
                    )

                    # Get supported sample formats
                    check_call(_lib.snd_pcm_format_mask_malloc, (byref(c_fmask_p),),
                               "Could not allocate memory for snd_pcm_format_mask_t.")
                    try:
                        check_call(_lib.snd_pcm_hw_params_get_format_mask, (c_params_p, c_fmask_p),
                                   "Could not get sample formats.")

                    except LibAsoundError as exc:
                        log.error(str(exc))
                    else:
                        formats = tuple(decode_format_mask(c_fmask_p))
                        log.debug("Sample formats: %s", ",".join(f[1] for f in formats))
                    finally:
                        _lib.snd_pcm_format_mask_free(c_fmask_p)

                    # Get supported period times
                    check_call(_lib.snd_pcm_hw_params_get_periods_min,
                               (c_params_p, byref(c_min), byref(c_dir)),
                               "Could not get minimum periods count.")

                    check_call(_lib.snd_pcm_hw_params_get_periods_max,
                               (c_params_p, byref(c_max), byref(c_dir)),
                               "Could not get minimum periods count.")

                    log.debug("Min/max periods count: (%i, %i)", c_min.value, c_max.value)
                    periods = (c_min.value, c_max.value)

                    # Get supported buffer sizes
                    check_call(_lib.snd_pcm_hw_params_get_buffer_size_min,
                               (c_params_p, byref(c_min_long)),
                               "Could not get minimum buffer time.")

                    check_call(_lib.snd_pcm_hw_params_get_buffer_size_max,
                               (c_params_p, byref(c_max_long)),
                               "Could not get minimum buffer time.")

                    log.debug("Min/max buffer time: (%i, %i) us", c_min_long.value,
                              c_max_long.value)
                    buffer_sizes = tuple(
                        size for size in PCM_BUFFER_SIZES
                        if c_min_long.value <= size <= c_max_long.value and
                        _lib.snd_pcm_hw_params_test_buffer_size(c_pcm_p, c_params_p, size) == 0
                    )

                    # List subdevices
                    for subd in range(0, nsubd):
                        _lib.snd_pcm_info_set_subdevice(c_pcminfo_p, c_int(subd))
                        sub_name = _lib.snd_pcm_info_get_subdevice_name(c_pcminfo_p).decode()
                        log.debug('Discovered subdevice: "%s"', sub_name)
                        subdevices.append(sub_name)
                except LibAsoundError as exc:
                    log.warning(exc)
                finally:
                    _lib.snd_pcm_close(c_pcm_p)
                    _lib.snd_pcm_hw_params_free(c_params_p)
                    _lib.snd_pcm_info_free(c_pcminfo_p)

            devices.append(
                AlsaDevice(
                    devno=c_dev.value,
                    id=device_id,
                    name=device_name,
                    stream=stream,
                    buffer_sizes=buffer_sizes,
                    periods=periods,
                    channels=channels,
                    rates=rates,
                    formats=formats,
                    subdevices=subdevices
                )
            )

        _lib.snd_ctl_close(c_handle_p)
        _lib.snd_ctl_card_info_free(c_info_p)

    return cards


class AlsaInfo:
    def __init__(self, deferred=True):
        if not deferred:
            _ = self.devices  # noqa

    def _make_device_list(self, cards):
        devs = []
        for card in cards:
            if not card.devices:
                continue

            devs.append('hw:%i' % card.cardno)
            devs.append('hw:%s' % card.id)

            for dev in card.devices:
                devs.append('hw:%i,%i' % (card.cardno, dev.devno))
                devs.append('hw:%s,%i' % (card.id, dev.devno))
                devs.append('hw:%s,%s' % (card.id, dev.id))

        return devs

    @property
    def playback_devices(self):
        cards = getattr(self, '_playback', None)
        if cards is None:
            self._playback = cards = get_cards(stream=SndPcmStream.PLAYBACK)
        return self._make_device_list(cards)

    @property
    def capture_devices(self):
        cards = getattr(self, '_capture', None)
        if cards is None:
            self._capture = cards = get_cards(stream=SndPcmStream.CAPTURE)
        return self._make_device_list(cards)

    @property
    def devices(self):
        devs = set(self.playback_devices)
        return list(devs.union(self.capture_devices))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG if '-v' in sys.argv[1:] else logging.INFO,
                        format="[%(name)s] %(levelname)s: %(message)s")

    ai = AlsaInfo()

    if '-r' in sys.argv:
        for card in get_cards(stream=SndPcmStream.CAPTURE):
            print(card)

        for card in ai.capture_devices:
            print(card)

        print("Capture devices")
        print("----------------\n")

    else:
        for card in get_cards(stream=SndPcmStream.PLAYBACK):
            print(card)

        print("Playback devices")
        print("----------------\n")

        for card in ai.playback_devices:
            print(card)



"""D-BUS service interface for jack-select."""

import logging
import os

import dbus
import dbus.service


log = logging.getLogger(__name__)

DBUS_NAME = 'de.chrisarndt.JackSelectService'
DBUS_PATH = '/de/chrisarndt/JackSelectApp'
DBUS_INTERFACE = 'de.chrisarndt.JackSelectInterface'


class JackSelectService(dbus.service.Object):
    def __init__(self, app, bus=None):
        if bus is None:
            bus = dbus.SessionBus()

        # we need to keep a reference to the BusName
        # otherwise it gets garbage-collected and the service vanishes
        self.bus_name = dbus.service.BusName(DBUS_NAME, bus)
        super().__init__(bus, DBUS_PATH)
        self.app = app

    @dbus.service.method(dbus_interface=DBUS_INTERFACE, out_signature='i')
    def GetPid(self):
        """Get the PID of the currently running JACK-Select process."""
        log.debug("DBus client requested PID.")
        return os.getpid()

    @dbus.service.method(dbus_interface=DBUS_INTERFACE)
    def Exit(self):
        """Exit JACK-Select."""
        log.debug("DBus client requested application exit.")
        self.app.quit()

    @dbus.service.method(dbus_interface=DBUS_INTERFACE)
    def OpenMenu(self):
        """Open the JACK-Select popup menu."""
        log.debug("DBus client requested opening menu.")
        self.app.open_menu()

    @dbus.service.method(dbus_interface=DBUS_INTERFACE, in_signature='s')
    def ActivatePreset(self, preset):
        """Activate the JACK configuration preset with the given name."""
        log.debug("DBus client requested activating preset '%s'." % preset)
        self.app.activate_preset(preset=preset)

    @dbus.service.method(dbus_interface=DBUS_INTERFACE)
    def ActivateDefaultPreset(self):
        """Activate the default JACK configuration preset."""
        log.debug("DBus client requested activating default preset.")
        self.app.activate_default_preset()

    @dbus.service.method(dbus_interface=DBUS_INTERFACE)
    def StopJackServer(self):
        """Stop the JACK server."""
        log.debug("DBus client requested stopping JACK server.")
        self.app.stop_jack_server()



# -*- coding: utf-8 -*-
"""Control and configure a aj2midid service via D-BUS."""

import logging

from .dbusinterface import DBUSBaseInterface


log = logging.getLogger(__name__)


class A2JCtlInterface(DBUSBaseInterface):
    service = "org.gna.home.a2jmidid"
    interface = "org.gna.home.a2jmidid.control"
    object_path = "/"

    # interface method wrappers

    def exit(self, cb=None):
        return self.call_async('exit', callback=cb)

    def is_started(self, cb=None):
        return self.call_async('is_started', callback=cb)

    def start(self, cb=None):
        return self.call_async('start', callback=cb)

    def stop(self, cb=None):
        return self.call_async('stop', callback=cb)

    def get_hw_export(self, cb=None):
        return self.call_async('get_hw_export', callback=cb)

    def set_hw_export(self, hw_export=True, cb=None):
        return self.call_async('set_hw_export', args=(hw_export,), callback=cb)

    def get_jack_client_name(self, cb=None):
        return self.call_async('get_jack_client_name', callback=cb)

    def get_disable_port_uniqueness(self, cb=None):
        return self.call_async('get_disable_port_uniqueness', callback=cb)

    def set_disable_port_uniqueness(self, disable_port_uniqueness=True, cb=None):
        return self.call_async('set_disable_port_uniqueness', args=(disable_port_uniqueness,),
                               callback=cb)

    def map_alsa_to_jack_port(self, alsa_client_id, alsa_port_id, map_playback, cb=None):
        return self.call_async('map_alsa_to_jack_port',
                               args=(alsa_client_id, alsa_port_id, map_playback),
                               callback=cb)

    def map_jack_port_to_alsa(self, jack_port_name, cb=None):
        return self.call_async('map_jack_port_to_alsa', args=(jack_port_name,), callback=cb)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""A systray app to set the JACK configuration from QjackCtl presets via DBus.

This application displays an icon in the system tray (also known as
notification area) of your desktop, which shows the status of the JACK_ audio
server and when you click on it, a menu pops up, which lets you quickly select
from the JACK configuration presets you created with QjackCtl_. When you
select a preset, its JACK engine and driver configuration settings are loaded
via DBus into JACK and then the server is restarted. This allows you to switch
between different audio setups with just two mouse clicks.

**jack-select** works with the DBus-version of JACK only. It is written in
Python 3 using the ``PyGObject`` bindings for GTK 3. Python 2 is not supported.

It is available from the source code repository on GitHub:

https://github.com/SpotlightKid/jack-select

Releases can be downloaded from the Python Package Index:

https://pypi.org/project/jack-select

**jack-select** is also available as an Arch Linux package from the Arch User
Repository:

https://aur.archlinux.org/packages/jack-select/


.. _jack: http://jackaudio.org/
.. _qjackctl: http://qjackctl.sourceforge.net/

"""

import setuptools


exec(open('jackselect/version.py').read())

setuptools.setup(
    name="jack-select",
    version=__version__,  # noqa
    url="https://github.com/SpotlightKid/jack-select",
    author="Christopher Arndt",
    author_email="chris@chrisarndt.de",
    description=__doc__.splitlines()[0],
    long_description="\n".join(__doc__.splitlines()[2:]),
    keywords="JACK,systray,GTK,DBus,audio",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
        'PyGObject',
        'dbus-python',
        'pyudev',
        'pyxdg'
    ],
    entry_points = {
        'console_scripts': [
            'jack-select = jackselect.jackselect:main',
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Environment :: X11 Applications :: GTK',
        'Topic :: Multimedia :: Sound/Audio'
    ],
)



# -*- coding: utf-8 -*-
"""Control and configure a JACK server via D-BUS."""

import logging

import dbus

from .dbusinterface import DBUSBaseInterface


log = logging.getLogger(__name__)

SETTINGS = {
    'engine': (
        ('client-timeout', dbus.Int32),
        ('clock-source', dbus.UInt32),
        ('driver', dbus.String),
        ('name', dbus.String),
        ('port-max', dbus.UInt32),
        ('realtime', dbus.Boolean),
        ('realtime-priority', dbus.Int32),
        ('replace-registry', dbus.Boolean),
        ('self-connect-mode', dbus.Byte),
        ('slave-drivers', dbus.String),
        ('sync', dbus.Boolean),
        ('temporary', dbus.Boolean),
        ('verbose', dbus.Boolean),
    ),
    'driver': (
        ('capture', dbus.String),
        ('device', dbus.String),
        ('dither', dbus.Byte),
        ('hwmeter', dbus.Boolean),
        ('hwmon', dbus.Boolean),
        ('inchannels', dbus.UInt32),
        ('midi-driver', dbus.String),
        ('monitor', dbus.Boolean),
        ('nperiods', dbus.UInt32),
        ('outchannels', dbus.UInt32),
        ('period', dbus.UInt32),
        ('playback', dbus.String),
        ('rate', dbus.UInt32),
        ('shorts', dbus.Boolean),
        ('softmode', dbus.Boolean),
    )
}


class JackBaseInterface(DBUSBaseInterface):
    service = "org.jackaudio.service"
    object_path = "/org/jackaudio/Controller"


class JackCtlInterface(JackBaseInterface):
    interface = "org.jackaudio.JackControl"

    def exit(self, cb=None, error_cb=None):
        return self.call_async('Exit', name='is_started', callback=cb,
                               error_callback=error_cb)

    def is_started(self, cb=None, error_cb=None):
        return self.call_async('IsStarted', name='is_started', callback=cb,
                               error_callback=error_cb)

    def is_realtime(self, cb=None, error_cb=None):
        return self.call_async('IsRealtime', name='is_realtime', callback=cb,
                               error_callback=error_cb)

    def start_server(self, cb=None, error_cb=None):
        return self.call_async('StartServer', name='start_server', callback=cb,
                               error_callback=error_cb)

    def stop_server(self, cb=None, error_cb=None):
        return self.call_async('StopServer', name='stop_server', callback=cb,
                               error_callback=error_cb)

    def get_latency(self, cb=None, error_cb=None):
        return self.call_async('GetLatency', name='latency', callback=cb,
                               error_callback=error_cb)

    def get_load(self, cb=None, error_cb=None):
        return self.call_async('GetLoad', name='load', callback=cb,
                               error_callback=error_cb)

    def get_period(self, cb=None, error_cb=None):
        return self.call_async('GetBufferSize', name='period', callback=cb,
                               error_callback=error_cb)

    def get_sample_rate(self, cb=None, error_cb=None):
        return self.call_async('GetSampleRate', name='samplerate', callback=cb,
                               error_callback=error_cb)

    def get_xruns(self, cb=None, error_cb=None):
        return self.call_async('GetXruns', name='xruns', callback=cb)

    def add_signal_handler(self, handler, signal=None):
        return self._if.connect_to_signal(
            signal_name=signal,
            handler_function=handler,
            interface_keyword='interface',
            member_keyword='signal')


class JackCfgInterface(JackBaseInterface):
    interface = "org.jackaudio.Configure"

    def engine_has_feature(self, feature):
        try:
            features = self._if.ReadContainer(["engine"])[1]
        except:  # noqa:E722
            features = ()
        return dbus.String(feature) in features

    def get_engine_parameter(self, parameter, fallback=None):
        if not self.engine_has_feature(parameter):
            return fallback
        else:
            try:
                return self._if.GetParameterValue(["engine", parameter])[2]
            except:  # noqa:E722
                return fallback

    def set_engine_parameter(self, parameter, value, optional=True):
        if not self.engine_has_feature(parameter):
            return 2
        elif optional:
            pvalue = self._if.GetParameterValue(["engine", parameter])

            if pvalue is None:
                return False

            if value != pvalue[2]:
                return bool(self._if.SetParameterValue(["engine", parameter],
                                                       value))
            else:
                return 3
        else:
            return bool(self._if.SetParameterValue(["engine", parameter],
                                                   value))

    def driver_has_feature(self, feature):
        try:
            features = self._if.ReadContainer(["driver"])[1]
        except:  # noqa:E722
            features = ()
        return dbus.String(feature) in features

    def get_driver_parameter(self, parameter, fallback=None):
        if not self.driver_has_feature(parameter):
            return fallback
        else:
            try:
                return self._if.GetParameterValue(["driver", parameter])[2]
            except:  # noqa:E722
                return fallback

    def set_driver_parameter(self, parameter, value, optional=True):
        if not self.driver_has_feature(parameter):
            return 2
        elif optional:
            if value != self._if.GetParameterValue(["driver", parameter])[2]:
                return bool(self._if.SetParameterValue(["driver", parameter],
                                                       value))
            else:
                return 3
        else:
            return bool(self._if.SetParameterValue(["driver", parameter],
                                                   value))

    def activate_preset(self, settings):
        for component in ('engine', 'driver'):
            csettings = settings.get(component, {})

            for setting in SETTINGS[component]:
                if isinstance(setting, tuple):
                    setting, stype = setting
                else:
                    stype = None

                value = csettings.get(setting)

                if value is None:
                    log.debug("Resetting %s.%s", component, setting)
                    self._if.ResetParameterValue([component, setting])
                    continue

                if stype:
                    dbus_value = stype(value)
                elif isinstance(value, bool):
                    dbus_value = dbus.Boolean(value)
                elif isinstance(value, int):
                    dbus_value = dbus.UInt32(value)
                elif isinstance(value, str):
                    dbus_value = dbus.String(value)
                else:
                    log.warning("Unknown type %s for setting '%s' = %r.",
                                type(value), setting, value)
                    dbus_value = value

                if component == 'engine':
                    setter = self.set_engine_parameter
                elif component == 'driver':
                    setter = self.set_driver_parameter

                log.debug("Setting %s.%s = %r", component, setting, value)
                result = setter(setting, dbus_value)

                if result not in (0, 3):
                    log.error("Setting %s setting '%s' failed (value %r), return value %s",
                              component, setting, value, result)



# -*- coding: utf-8 -*-
"""Read JACK presets and their settings from QjackCtl's configuration file."""

import configparser
import logging


log = logging.getLogger(__name__)
PARAM_MAPPING = {
    'chan': ('driver', ('inchannels', 'outchannels')),
    'clocksource': 'clock-source',
    'driver': ('engine', 'driver'),
    'frames': 'period',
    'indevice': 'capture',
    'inlatency': 'input-latency',
    'interface': 'device',
    'mididriver': 'midi-driver',
    'outdevice': 'playback',
    'outlatency': 'output-latency',
    'periods': 'nperiods',
    'portmax': ('engine', 'port-max'),
    'priority': ('engine', 'realtime-priority'),
    'realtime': ('engine', 'realtime'),
    'samplerate': 'rate',
    'selfconnectmode': 'self-connect-mode',
    'servername': ('engine', 'name'),
    'slavedrivers': ('engine', 'slave-drivers'),
    'timeout': ('engine', 'client-timeout'),
    'verbose': ('engine', 'verbose'),
    # 'snoop': '???'
}
VALUE_MAPPING = {
    'clock-source': {
        'c': 0,
        'h': 1,
        's': 2,
    },
    'dither': {
        '0': b'n',
        '1': b'r',
        '2': b's',
        '3': b't',
    },
    'self-connect-mode': {
        '0': b' ',
        '1': b'e',
        '2': b'E',
        '3': b'a',
        '4': b'A',
    }
}
ALLOWED_VALUES = {
    'clock-source': VALUE_MAPPING['clock-source'].values(),
    'dither': VALUE_MAPPING['dither'].values(),
    'midi-driver': ('none', 'seq', 'raw'),
    'self-connect-mode': VALUE_MAPPING['self-connect-mode'].values(),
}
DEFAULT_PRESET = '(default)'


def get_qjackctl_presets(qjackctl_conf, ignore_default=False):
    config = configparser.ConfigParser()
    config.optionxform = lambda option: option
    config.read(qjackctl_conf)

    preset_names = set()
    settings = {}

    if 'Settings' in config:
        for name in sorted(config['Settings']):
            try:
                preset_name, param = name.split('\\', 1)
            except ValueError:
                # The default (nameless) preset was saved.
                # It uses settings keys without a preset name prefix.
                param = name
                preset_name = DEFAULT_PRESET

            preset_names.add(preset_name)
            param = param.lower()
            value = config.get('Settings', name)
            param = PARAM_MAPPING.get(param, param)

            if isinstance(param, tuple):
                component, param = param
            else:
                component = 'driver'

            if preset_name not in settings:
                settings[preset_name] = {}

            if component not in settings[preset_name]:
                settings[preset_name][component] = {}

            if not isinstance(param, (tuple, list)):
                param = (param,)

            for p in param:
                if p in VALUE_MAPPING:
                    value = VALUE_MAPPING[p].get(value, value)
                if value == 'false':
                    value = False
                elif value == 'true':
                    value = True
                elif p in ALLOWED_VALUES and value not in ALLOWED_VALUES[p]:
                    value = None
                elif value == '':
                    value = None
                else:
                    try:
                        value = int(value)
                    except (TypeError, ValueError):
                        pass

                settings[preset_name][component][p] = value

    if (ignore_default and DEFAULT_PRESET in settings and len(settings) > 1):
        del settings[DEFAULT_PRESET]

        if DEFAULT_PRESET in preset_names:
            preset_names.remove(DEFAULT_PRESET)

    default_preset = config.get('Presets', 'DefPreset', fallback=None)

    if default_preset not in preset_names:
        default_preset = None

    if not default_preset and DEFAULT_PRESET in preset_names:
        default_preset = DEFAULT_PRESET

    return list(preset_names), settings, default_preset


def _test():
    from xdg import BaseDirectory as xdgbase

    qjackctl_conf = xdgbase.load_first_config('rncbc.org/QjackCtl.conf')

    if qjackctl_conf:
        presets, settings, default = get_qjackctl_presets(qjackctl_conf, True)
        for preset in sorted(presets):
            print("%s:%s\n" % (preset, " *" if preset == default else ''))
            print("[engine]")

            for setting, value in sorted(settings[preset]['engine'].items()):
                print("%s: %r" % (setting, value))

            print("\n[driver]")

            for setting, value in sorted(settings[preset]['driver'].items()):
                print("%s: %r" % (setting, value))

            print()


if __name__ == '__main__':
    _test()
